"""Provide fixtures for the entire parent directory."""

from tests.fixtures.slp import *
from tests.fixtures.labels import *
from tests.fixtures.labelstudio import *
from tests.fixtures.videos import *
from tests.fixtures.jabs import *
